# ImageUploadUtility - (c) 2012-2017 Juxtaposition. All Rights Reserved.
# This code cannot be redistributed without permission from juxtaposition.jp
# For more information, consult your ImageUploadUtility license.
#

package ImageUploadUtility::Plugin;
use strict;
use warnings;
use utf8;
use Carp;
use Data::Dumper;
use MT::Util qw( perl_sha1_digest_hex );

use constant DO_NOT_CONVERT     => 0;
use constant CONVERT_MULTI_BYTE => 1;
use constant FORCE_CONVERT      => 2;
use constant RESIZE_DISABLED    => 0;
use constant RESIZE_LONGSIDE    => 1;
use constant RESIZE_WIDTH       => 2;

use constant DEBUG => 0;

sub _trace {
    my $msg = shift;
    my @info = caller(1);
    print STDERR "TRACE: $info[0]: $info[3]: $msg\n" if DEBUG;
}

sub _t {
    my $msg = shift;
    return unless DEBUG;
    require Time::HiRes;
    require File::Spec;
    my $app = MT->instance();
    my $plugin = $app->component('ImageUploadUtility');
    my $logfile = File::Spec->catfile($plugin->path, 'data', 'trace.log');
    local $| = 1;
    open LOG, ">>:encoding(UTF-8)", $logfile or die $!;
    my $now = Time::HiRes::time();
    my @info = caller(1);
    print LOG "TRACE:[$$]:[$now]: " . $info[0] . ": " . $info[3] . ": " . $msg . "\n";
    close LOG;
}

my $sub_replaced = 0;

sub init_app {
    my ($cb, $app) = @_;

    return 1 if $sub_replaced;

    my $plugin = $app->component('ImageUploadUtility')
        or croak "Plugin not found.";
    my $has_community = eval { require MT::App::Community; 1 } ? 1 : 0;

    # Resize uploaded image
    require MT::CMS::Asset;
    no strict 'refs';
    no warnings 'redefine';
    my $saved_upload_file = \&MT::CMS::Asset::upload_file;
    my $saved_js_upload_file = \&MT::CMS::Asset::js_upload_file if exists &MT::CMS::Asset::js_upload_file;
    my $saved_handle_upload = $has_community ? \&MT::App::Community::_handle_upload : undef;
    my $_upload_file = sub {
        my $app = shift;
        my $blog_id = $app->param('blog_id') || 0;
        my $edit_field = $app->param('edit_field') || '';
        my $enable_resize = defined $app->param('enable_resize') 
            ? $app->param('enable_resize') : $plugin->enable_resize($blog_id);
        my $watermark_asset_id = $plugin->watermark_asset_id($blog_id);
        my $insert_watermark = $watermark_asset_id ? 1 : 0;
        my $strip_exif = $plugin->strip_exif($blog_id);
        my $jpeg_quality = $plugin->jpeg_quality($blog_id);
        my $normalize_orientation = $plugin->normalize_orientation($blog_id);
        my $js_upload = !$plugin->is_upload_compat;
        my $return_func = $js_upload ? $saved_js_upload_file : $saved_upload_file;
        my $dialog = $app->param('dialog') || 0;
        # アイテムの新規作成メニューからリサイズする用
        my $resize_item = $app->param('image_upload_utility_resize_item') || 0;

        unless($enable_resize || $strip_exif || $jpeg_quality || $insert_watermark || $normalize_orientation) {
            if (ref $app eq 'MT::App::CMS') {
                return $return_func->($app, @_);
            } elsif ($has_community && ref $app eq 'MT::App::Community') {
                return $saved_handle_upload->($app, @_);
            } else {
                croak 'Unknown app.';
            }
        }

        my $jpeg_quality_changed = 0;

        require MT::Image;
        no strict 'refs';
        no warnings 'redefine';
        my $saved_check_upload = \&MT::Image::check_upload;
        local *MT::Image::check_upload = sub {
            my $class = shift;
            my %params = @_;
            my ($w, $h, $id, $write_file) = $saved_check_upload->($class, @_);

            # saved_check_uploadはpdfをhandlingできないので、w/h/idをここで書き換える
            if($plugin->has_extra && $plugin->enable_pdf_thumbnail) {
                require File::Basename;
                my $localfile = $params{Local} || '';
                my ( $filename, $path, $ext )
                    = ( File::Basename::fileparse( $localfile, qr/[A-Za-z0-9]+$/ ) );
                if($id ne 'pdf' && $ext eq 'pdf' 
                    && !defined $w && !defined $h && $write_file) {
                    ($w, $h) = ImageUploadUtility::Plugin::Extra::get_size(Fh => $params{Fh});
                    if(!defined $w) {
                        return (undef, undef, 'Data stream is not a known image file format', $write_file);
                    }
                    return ($w, $h, 'pdf', $write_file);
                }
            }

            unless($w && $h && $write_file
                       && $id =~ m!^(?:jpg|gif|png)$!i
                           && ($enable_resize || $strip_exif 
                               || $jpeg_quality || $insert_watermark || $normalize_orientation)) {
                return ($w, $h, $id, $write_file);
            }

            my $fh = $params{Fh};
            seek $fh, 0, 0;
            my $uploaded_data = eval { local $/; <$fh> };
            seek $fh, 0, 0;

            # normalize orientation
            if ($normalize_orientation) {
                ($uploaded_data, $w, $h) 
                    = normalize_orientation($uploaded_data, $w, $h, lc($id));
            }

            my $img = $class->new( Data => $uploaded_data, Type => lc($id) )
                or return $class->error($class->errstr);
            my $magick = $img->can('_magick') ? $img->_magick : (exists $img->{magick} ? $img->{magick} : undef);

            # resize image
            my $resized_data;
            if ($dialog || $resize_item) {
                my ($max_size, $long_side);
                if ($enable_resize == RESIZE_LONGSIDE) {
                    $max_size = defined $app->param('max_size') ? $app->param('max_size') : $plugin->max_size($blog_id);
                    $long_side = $w >= $h ? 'Width' : 'Height';
                }
                elsif ($enable_resize == RESIZE_WIDTH) {
                    $max_size = defined $app->param('max_width') ? $app->param('max_width') : $plugin->max_width($blog_id);
                    $long_side = 'Width';
                }
                if ($enable_resize && $max_size
                    && (($long_side eq 'Width' && defined($w) && $w > $max_size)
                    || ($long_side eq 'Height' && defined($h) && $h > $max_size))) {
                    # NOTE: scale method will remove exif info.
                    local $MT::L10N::ja::Lexicon{'Scaling to [_1]x[_2] failed: [_3]'} = 'サイズを[_1]x[_2]に変更できませんでした: [_3]';
                    ( $resized_data, $w, $h ) = $img->scale( $long_side => $max_size );
                    $app->log('[ImageUploadUtility] '
                        . $plugin->translate('File [_1] scaled to [_2]x[_3] ([_4] bytes).',
                            $app->param('file'), $w, $h, length($resized_data)));
                    if($img->errstr || !$resized_data) {
                        $app->log('[ImageUploadUtility] ' . $img->errstr . ', ' . $app->param('file'));
                        return $class->error( $img->errstr );
                    }
                    if($magick) {
                        my ($_w, $_h) = $magick->Get('width', 'height');
                        if($w != $_w || $h != $_h) {
                            my $msg = $plugin->translate('Scaling to [_1]x[_2] failed: [_3]', $w, $h, $plugin->translate('Failed to resize.'));
                            $app->log('[ImageUploadUtility] ' . $msg . ', ' . $app->param('file'));
                            return $class->error($msg);
                        }
                    }
                    _log('resize image: 1');
                }
            }
    
            # add watermark
            if ($plugin->has_extra && $magick && $insert_watermark) {
                if (ImageUploadUtility::Plugin::Extra::add_watermark($app, $magick, $watermark_asset_id)) {
                    $resized_data = undef;
                    $app->log('[ImageUploadUtility] '
                        . $plugin->translate('File [_1] inserted watermark (ID=[_2]).',
                            $app->param('file'), $watermark_asset_id));
                    _log('insert watermark: 1');
                }
            }

            if($magick) {
                if ( $strip_exif ) {
                    $magick->Strip() if $magick->can('Strip');
                    $app->log('[ImageUploadUtility] '
                        . $plugin->translate('File [_1] deleted Exif info.', $app->param('file')));
                }
                if ( $magick && $id =~ m!^jpg$!i && $jpeg_quality ) {
                    $resized_data = $img->blob($jpeg_quality);
                    $app->log('[ImageUploadUtility] '
                        . $plugin->translate('File [_1] changed quality to [_2] ([_3] bytes).',
                            $app->param('file'), $jpeg_quality, length($resized_data)));
                    $jpeg_quality_changed = 1;
                }
            }

            if ( !$resized_data ) {
                if ( !$magick ) {
                    return ( $w, $h, $id, $write_file );
                }
                else {
                    $resized_data = $magick->ImageToBlob;
                }
            }

            my $file_size = length $resized_data;
            my $max_filesize = $params{Max};
            if ($max_filesize && $max_filesize < $file_size) {
                return $class->error(
                    MT->translate(
                        "File size exceeds maximum allowed: [_1] > [_2]",
                        $file_size, $max_filesize,
                    )
                );
            }

            return (
                $w, $h, $id,
                sub {
                    $params{Fmgr}->put_data($resized_data, $params{Local}, 'upload');
                },
            );
        };

        # HACK: is_protectedなPDFのiconを置き換える
        my $saved_json_result = \&MT::App::json_result;
        local *MT::App::json_result = sub {
            my $app = shift;
            my ($result) = @_;
            if($plugin->has_extra) {
                $result = ImageUploadUtility::Plugin::Extra::replace_protected_pdf_icon($result);
            }
            return $saved_json_result->($app, $result);
        } if $js_upload && $plugin->has_extra;

        # HACK: JPEG画像品質の自動変換を無効にする
        require MT::Asset::Image;
        my $saved_change_quality = \&MT::Asset::Image::change_quality;
        local *MT::Asset::Image::change_quality = sub {
            my (@args) = @_;
            # jpeg_quality変換済みの場合はskipする
            return $jpeg_quality_changed ? 1 : $saved_change_quality->(@_);
        };

        if (ref $app eq 'MT::App::CMS') {
            my ($asset, $bytes) = $return_func->($app, @_);
            if($plugin->has_extra) {
                $asset = ImageUploadUtility::Plugin::Extra::rebless_asset_if_pdf($asset);
            }
            return ($asset, $bytes);
        } elsif ($has_community && ref $app eq 'MT::App::Community') {
            return $saved_handle_upload->($app, @_);
        } else {
            croak 'Unknown app.';
        }
    };
    *MT::CMS::Asset::upload_file = $_upload_file;
    *MT::CMS::Asset::js_upload_file = $_upload_file if $saved_js_upload_file;
    *MT::App::Community::_handle_upload = $_upload_file if $has_community;

    # Convert multibyte filename
    require MT::App;
    my $saved_upload_info = \&MT::App::upload_info;
    *MT::App::upload_info = sub {
        my $app = shift;
        my ($param_name) = @_;
        # default parameter name for uploaded file.
        $param_name ||= 'file';
        my $blog_id = $app->param('blog_id') || 0;
        my ($fh, $info) = $saved_upload_info->($app, @_);
        my $enable_rename = $plugin->enable_rename($blog_id);
        my $ext_lowercase = $plugin->ext_lowercase($blog_id);
        return ($fh, $info) unless $enable_rename || $ext_lowercase;
        return unless $fh;

        my $filename = $app->param($param_name) || $app->param('fname');
        return ($fh, $info) unless $filename;
        $filename =~ s!\\!/!g;
        $filename =~ s!^.*/!!;
        my @parts = File::Basename::fileparse($filename, qr/[A-Za-z0-9]+$/);
        my $ext = $parts[2];

        if($filename && $filename !~ m!\.\.|\0|\|!) {
            if(($enable_rename == CONVERT_MULTI_BYTE && $filename =~ m/[^\x20-\x7E]/)
                || $enable_rename == FORCE_CONVERT) {
                my $format = $plugin->rename_format($blog_id);
                require MT::Template;
                my $tmpl = MT::Template->new(
                    type   => 'scalarref',
                    source => \$format,
                );
                my $random_number = perl_sha1_digest_hex($app->make_magic_token);
                my $path = $tmpl->output({
                    random_number => $random_number,
                });
                $path = time() unless $path;
                $ext = lc $ext if $ext_lowercase && $ext;
                $path .= '.' . $ext if $ext;

                $app->param('fname', $path);
                $app->param($param_name, $path);
            }
            elsif ($ext_lowercase && $ext) {
                my $path = $parts[0] . lc $ext;
                $app->param('fname', $path);
                $app->param($param_name, $path);
            }
        }
        return ($fh, $info);
    };

    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::init_app($cb, $app);
    }

    $sub_replaced = 1;

    return 1;
}

use POSIX qw(floor);

sub _adjust_size {
    my ($w, $h, $fw, $fh, $max) = @_;
    my $dir = $w >= $h ? 'w' : 'h';
    my $ratio;
    if($dir eq 'w' && $w > $max) {
        $w = $max;
        $ratio = $w / $fw;
        $h = floor($fh * $ratio);
    }
    elsif($dir eq 'h' && $h > $max) {
        $h = $max;
        $ratio = $h / $fh;
        $w = floor($fw * $ratio);
    }
    return ($w, $h);
}

sub _log {
    MT->component('ImageUploadUtility')->doLog(@_) if DEBUG;
}

sub fltr_image_max_size {
    my ($str, $val, $ctx) = @_;
    my $max_size = $val;
    return $str unless $max_size;

    $str =~ s!
                 ( # $1 = whole tag
                     <(?:[Ii][Mm][Gg])
                     ([^>]+) # $2 = pre attr
                     src="([^"]+)" # $3 = src
                     ([^>]*) # $4 = post attr
                     \/?>
                 )
             !
                 resize_image($ctx, $max_size, $1, $3, "$2 $4")
             !gxmse;

    return $str;
}

sub fltr_image_max_width {
    my ($str, $val, $ctx) = @_;
    my $max_width = $val;
    return $str unless $max_width;

    $str =~ s!
                 ( # $1 = whole tag
                     <(?:[Ii][Mm][Gg])
                     ([^>]+) # $2 = pre attr
                     src="([^"]+)" # $3 = src
                     ([^>]*) # $4 = post attr
                     \/?>
                 )
             !
                 resize_image($ctx, $max_width, $1, $3, "$2 $4", 'width')
             !gxmse;

    return $str;
}

sub resize_image {
    my ($ctx, $max_size, $tag, $src, $attr, $long_side) = @_;
    return $tag unless $tag && $src && $attr;
    return $tag unless $max_size;
    return $tag unless $attr =~ m!class="mt-image-(none|left|right|center)"!;
    my $align = $1;

    require URI;
    my $uri = URI->new($src);
    my @paths = $uri->path_segments;
    my $file = pop @paths
        or return $tag;
    my ($base, $path, $ext) =
        File::Basename::fileparse($file, qr/[A-Za-z0-9]+$/);
    return $tag unless $ext =~ m!(?:jpe?g|gif|png)!i;
    my $filename = $base.$ext;

    require MT::Asset;
    my $asset;
    my @assets = MT::Asset->load(
        {
            file_name => $filename,
            class => 'image',
        },
    ) or return $tag;
    foreach my $a (@assets) {
        if($src eq $a->url) {
            $asset = $a;
            last;
        }
    }
    return $tag unless $asset;

    if(_need_resize($asset, $max_size, $long_side)) {
        if($asset->parent) {
            my $parent = MT::Asset->load($asset->parent)
                or return $tag;
            return $tag unless _need_resize($parent, $max_size, $long_side);
            $asset = $parent;
        }
        my ($max_w, $max_h) = _max_dimension($asset, $max_size, $long_side);
        my $param = {
            id => $asset->id,
            new_entry => 1,
            thumb => 1,
            thumb_width => $max_w,
            thumb_height => $max_h,
            align => $align,
            include => 1,
            wrap_text => 1,
        };
        my $html = create_thumbnail_link($asset, $param)
            or return $tag;
        $html =~ s!(?:<a[^>]+>)?(<img[^>]+\/>)(?:<\/a>)?!$1!;
        return $html;
    }

    return $tag;
}

sub create_thumbnail_link {
    my ($asset, $param) = @_;

    my $app = MT->instance;
    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 

    my $reblessed = 0;
    if(ref $app eq 'MT') {
        bless $app, 'MT::App';
        die 'Failed to rebless to MT::App' unless $app->can('user');
        my $user = $asset->created_by ? MT::Author->load($asset->created_by) : MT::Author->new;
        $app->user($user);
        require CGI;
        $app->{query} = CGI->new;
        MT->set_instance($app);
        $reblessed = 1;
    }

    my $code = sub {
        if($plugin->has_extra && $plugin->enable_pdf_thumbnail) {
            my $ext = $asset->file_ext || '';
            if($ext eq 'pdf') {
                return ImageUploadUtility::Plugin::Extra::create_thumbnail_link($asset, $param);
            }
        }
        $asset->on_upload($param)
            or return;
        my $html = $asset->as_html($param)
            or return;

        return $html;
    };

    my $html = eval { $code->(); };

    if($reblessed) {
        bless $app, 'MT';
        MT->set_instance($app);
    }
    return $html;
}

sub _need_resize {
    my ($asset, $max_size, $long_side) = @_;
    my ($w, $h) = _asset_dimension($asset);
    unless(defined $long_side && $long_side =~ m!^(?:width|height)$!) {
        $long_side = $w >= $h ? 'width' : 'height';
    }
    if ($max_size
            && (($long_side eq 'width' && defined($w) && $w > $max_size)
                    || ($long_side eq 'height' && defined($h) && $h > $max_size))) {
        return 1;
    }
    return;
}

sub _max_dimension {
    my ($asset, $max_size, $long_side) = @_;
    my ($w, $h) = _asset_dimension($asset);
    unless(defined $long_side && $long_side =~ m!^(?:width|height)$!) {
        $long_side = $w >= $h ? 'width' : 'height';
    }
    my ($max_w, $max_h) = $long_side eq 'width'
        ? ($max_size, 0) : (0, $max_size);
    return ($max_w, $max_h);
}

sub _asset_dimension {
    my ($asset) = @_;
    my $app = MT->instance;
    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 
    if($plugin->has_extra && $plugin->enable_pdf_thumbnail) {
        my $ext = $asset->file_ext || '';
        if($ext eq 'pdf') {
            return ImageUploadUtility::Plugin::Extra::_asset_dimension($asset);
        }
    }
    my $meta = $asset->metadata;
    return ($meta->{image_width}, $meta->{image_height});
}

sub normalize_orientation {
    my ($data, $width, $height, $ext) = @_;
    return ($data, $width, $height) unless $data;
    require Image::ExifTool; 
    my $tool = Image::ExifTool->new 
        or return ($data, $width, $height);
    $tool->ExtractInfo(\$data);
    my $orientation = $tool->GetInfo('Orientation')->{'Orientation'};
    if ($orientation && $orientation ne 'Horizontal (normal)' && $orientation !~ /^Unknown/i) {
        $tool->SetNewValue(Orientation => $orientation, DelValue => 1);
        $tool->WriteInfo(\$data);
        my $image = MT::Image->new(Data => $data, Type => $ext)
            or return ($data, $width, $height);
        my ($new_data, $new_width, $new_height) = do {
            if ($orientation eq 'Mirror horizontal') {
                $image->flipVertical();
            }
            elsif ($orientation eq 'Rotate 180') {
                $image->rotate(Degrees => 180);
            }
            elsif ($orientation eq 'Mirror vertical') {
                $image->flipHorizontal();
            }
            elsif ($orientation eq 'Mirror horizontal and rotate 270 CW') {
                $image->flipVertical();
                $image->rotate(Degrees=> 270);
            }
            elsif ($orientation eq 'Rotate 90 CW') {
                $image->rotate(Degrees => 90);
            }
            elsif ($orientation eq 'Mirror horizontal and rotate 90 CW') {
                $image->flipVertical();
                $image->rotate(Degrees => 90);
            }
            elsif ($orientation eq 'Rotate 270 CW') {
                $image->rotate(Degrees => 270);
            }
        };
        return ($new_data, $new_width, $new_height) if $new_data;
    }
    return ($data, $width, $height);
}

sub edit_entry {
    my ($cb, $app, $tmpl_ref) = @_;

    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 
    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::multiupload_script($plugin, $app, $tmpl_ref);
    }

    return 1;
}

sub edit_param {
    my ($cb, $app, $param, $tmpl) = @_;

    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 
    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::multiupload_params($plugin, $app, $param, $tmpl);
    }
    return 1;
}

sub source_async_asset_upload {
    my ($cb, $app, $tmpl_ref, $file) = @_;
    
    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 
    
    # NOTE: Insert only Asset Upload Screen.
    if($app->mode eq 'start_upload') {
        # Insert resize option
        my $pattern = quotemeta(<<'HTML');
</form>
HTML
        my $replacement = <<'HTML';
<__trans_section component="ImageUploadUtility">
<mtapp:setting
    id="image_upload_utility_resize_item"
    label_class="left-label"
    label="<__trans phrase="Resizing image" />"
    hint="<__trans phrase="Settings such as maximum size follow the ImageUploadUtility plugin setting." />"
    show_hint="1">
    <input type="checkbox" id="image_upload_utility_resize_item" name="image_upload_utility_resize_item" value="1" />
    <label for="image_upload_utility_resize_item"><__trans phrase="Resize images when uploaded." /></label>
</mtapp:setting>
</__trans_section>
HTML
        $$tmpl_ref =~ s!($pattern)!$replacement$1!;
   
        # Insert resize param into JS
        my $js_pattern = quotemeta(<<'SCRIPT');
    return fd;
SCRIPT
        my $js_replacement = <<'SCRIPT';
  var resizeItem = 0;
  if ( jQuery('#image_upload_utility_resize_item').prop('checked') ) {
    resizeItem = 1;
  }
  fd.append('image_upload_utility_resize_item', resizeItem);

SCRIPT
        $$tmpl_ref =~ s!($js_pattern)!$js_replacement$1!;
    }

    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::source_async_asset_upload(@_);
    }

    return 1;
}

sub param_async_asset_upload {
    my ($cb, $app, $param, $tmpl) = @_;

    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 
    my $blog_id = $app->param('blog_id') || 0;
  
    my $rename_node = $tmpl->getElementById('auto_rename_non_ascii');
    if($rename_node && $plugin->enable_rename($blog_id)) {
        $rename_node->innerHTML(<<'TMPL');
<__trans_section component="ImageUploadUtility">
      <label for="auto_rename_non_ascii"><__trans phrase="The ImageUploadUtility setting takes precedence over this setting."></label>
</__trans_section>
TMPL
    }
    
    my $normalize_orientation_node = $tmpl->getElementById('normalize_orientation');
    if($normalize_orientation_node && $plugin->normalize_orientation($blog_id)) {
        $normalize_orientation_node->innerHTML(<<'TMPL');
<__trans_section component="ImageUploadUtility">
      <label for="normalize_orientation"><__trans phrase="The ImageUploadUtility setting takes precedence over this setting."></label>
</__trans_section>
TMPL
    }

    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::param_async_asset_upload(@_);
    }

    return 1;
}

sub param_asset_modal {
    my ($cb, $app, $param, $tmpl) = @_;

    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 

    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::param_asset_modal(@_);
    }

    return 1;
}

sub param_asset_upload {
    my ($cb, $app, $param, $tmpl) = @_;

    my $plugin = $app->component('ImageUploadUtility')
        or croak 'Plugin not found.'; 

    if($plugin->has_extra) {
        ImageUploadUtility::Plugin::Extra::param_asset_upload(@_);
    }

    return 1;
}

1;
__END__
